#include <QtCore/QString>
#include <QtTest/QtTest>
#include <QSharedPointer>
#include <tr1/memory>

class ObjetTest
{
public:
	ObjetTest() : ptr(0)
	{
		ptr = new int();
		_list.append (value);
		_list.append (*ptr);
	}

	~ObjetTest()
	{
		delete ptr;
	}

	int value, *ptr;
	QList<int> _list;
};

QQueue<ObjetTest*> _queue;

void returnToPool(ObjetTest *obj)
{
	_queue.enqueue(obj);
}

QSharedPointer<ObjetTest> createFromPool()
{
	ObjetTest * c_ptr;
	if (_queue.size())
	{
	  c_ptr = _queue.dequeue();
	}
	else
	{
	  c_ptr = new ObjetTest();
	}

	QSharedPointer<ObjetTest> ptr(c_ptr, returnToPool);
	return ptr;
}

std::tr1::shared_ptr<ObjetTest> createFromBoostPool()
{
	ObjetTest * c_ptr;
	if (_queue.size())
	{
	  c_ptr = _queue.dequeue();
	}
	else
	{
	  c_ptr = new ObjetTest();
	}

	std::tr1::shared_ptr<ObjetTest> ptr(c_ptr, returnToPool);
	return ptr;
}

/* SmartBenchTest */

class SmartBenchTest : public QObject
{
	Q_OBJECT
public:
	SmartBenchTest();

	int iterations(QString test_name, int test_number);
private Q_SLOTS:
	void testAllocation();
	void testAllocation_data();

	void testModification();
	void testModification_data();

	void testAffectation();
	void testAffectation_data();

	void testClear();
	void testClear_data();
private:
	int iterations_cnt;
	QString test_name;
	int test_number;
};

SmartBenchTest::SmartBenchTest()
{
	_queue.clear ();
}

int SmartBenchTest::iterations(QString test_name, int test_number)
{
	if ((test_name != this->test_name) || (test_number != this->test_number))
	{
		this->test_name = test_name;
		this->test_number = test_number;
		this->iterations_cnt = 1;
		QStringList args(qApp->arguments ());
		QStringList::const_iterator it = qFind(args, "-iterations");
		if ((it != args.constEnd ()) && (++it != args.constEnd ()))
		{
			bool ok;
			int par_iterations_cnt = (*it).toInt(&ok);
			if (ok)
			{
				this->iterations_cnt = par_iterations_cnt;
			}
		}
	}
	else
	{
		iterations_cnt *= 2;
	}
	return iterations_cnt;
}

void SmartBenchTest::testAllocation_data()
{
	QTest::addColumn<int>("use_smart");
	QTest::newRow("C Pointer") << 0;
	QTest::newRow("Qt Smart Pointer") << 1;
	QTest::newRow("Qt Smart Pointer as Pool") << 2;
	QTest::newRow("C++0x Smart Pointer") << 3;
	QTest::newRow("C++0x Smart Pointer as Pool") << 4;
}

void SmartBenchTest::testAllocation()
{
	QFETCH(int, use_smart);
	switch (use_smart)
	{
	case 0 :
		{
			QList<ObjetTest*> c_ptr_list;
			QBENCHMARK {
				ObjetTest* ptr = new ObjetTest();
				c_ptr_list.append (ptr);
			}
			qDeleteAll(c_ptr_list);
			c_ptr_list.clear();
		}
		break;
	case 1 :
		{
			QList< QSharedPointer<ObjetTest> > smart_ptr_list;
			QBENCHMARK
			{
				QSharedPointer<ObjetTest> ptr(new ObjetTest());
				smart_ptr_list.append (ptr);
			}
			smart_ptr_list.clear();
		}
		break;
	case 2 :
		{
			int nbi = iterations (QTest::currentTestFunction (), 2);
			for(int i = 0; i < nbi; i++)
			{
				_queue.enqueue (new ObjetTest());
			}

			{
				QList< QSharedPointer<ObjetTest> > pool_ptr_list;
				QBENCHMARK
				{
					QSharedPointer<ObjetTest> ptr = createFromPool ();
					pool_ptr_list.append (ptr);
				}
				pool_ptr_list.clear();
			}

			qDeleteAll(_queue);
			_queue.clear ();
		}

		break;
	case 3:
		{
			QList< std::tr1::shared_ptr<ObjetTest> > boost_ptr_list;
			QBENCHMARK
			{
				std::tr1::shared_ptr<ObjetTest> ptr(new ObjetTest());
				boost_ptr_list.append (ptr);
			}
			boost_ptr_list.clear();
		}
		break;
	case 4 :
		{
			int nbi = iterations (QTest::currentTestFunction (), 2);
			for(int i = 0; i < nbi; i++)
			{
				_queue.enqueue (new ObjetTest());
			}

			{
				QList< std::tr1::shared_ptr<ObjetTest> > pool_ptr_list;
				QBENCHMARK
				{
					std::tr1::shared_ptr<ObjetTest> ptr = createFromBoostPool ();
					pool_ptr_list.append (ptr);
				}
				pool_ptr_list.clear();
			}

			qDeleteAll(_queue);
			_queue.clear ();
		}

		break;
	}
}

void SmartBenchTest::testModification_data()
{
	QTest::addColumn<int>("use_smart");
	QTest::newRow("C Pointer") << 0;
	QTest::newRow("Qt Smart Pointer") << 1;
	QTest::newRow("C++0x Smart Pointer") << 2;
}

void SmartBenchTest::testModification()
{
	QFETCH(int, use_smart);
	int random_number = qrand ();

	switch (use_smart)
	{
	case 0:
		{
			ObjetTest * obj = new ObjetTest;
			QBENCHMARK {
				obj->value = random_number;
			}
		}
		break;
	case 1:
		{
			QSharedPointer<ObjetTest> obj(new ObjetTest);

			QBENCHMARK
			{
				obj->value = random_number;
			}
		}
		break;
	case 2:
		{
			std::tr1::shared_ptr<ObjetTest> obj(new ObjetTest);

			QBENCHMARK
			{
				obj->value = random_number;
			}
		}
		break;
	}
}

void SmartBenchTest::testAffectation_data()
{
	QTest::addColumn<int>("use_smart");
	QTest::newRow("C Pointer") << 0;
	QTest::newRow("Qt Smart Pointer") << 1;
	QTest::newRow("C++0x Smart Pointer") << 2;
}

void SmartBenchTest::testAffectation()
{
	QFETCH(int, use_smart);
	int random_number = qrand ();

	switch (use_smart)
	{
	case 0:
		{
			ObjetTest * obj = new ObjetTest;
			QBENCHMARK
			{
				ObjetTest * obj2 = obj;
				obj2->value = random_number;
			}
		}
		break;
	case 1:
		{
			QSharedPointer<ObjetTest> obj(new ObjetTest);

			QBENCHMARK
			{
				QSharedPointer<ObjetTest> obj2 = obj;
				obj2->value = random_number;
			}
		}
		break;
	case 2:
		{
			std::tr1::shared_ptr<ObjetTest> obj(new ObjetTest);

			QBENCHMARK
			{
				std::tr1::shared_ptr<ObjetTest> obj2 = obj;
				obj2->value = random_number;
			}
		}
		break;
	}
}

void SmartBenchTest::testClear_data()
{
	QTest::addColumn<int>("use_smart");
	QTest::newRow("C Pointer") << 0;
	QTest::newRow("Qt Smart Pointer") << 1;
	QTest::newRow("Qt Smart Pointer as Pool") << 2;
	QTest::newRow("C++0x Smart Pointer") << 3;
	QTest::newRow("C++0x Smart Pointer as Pool") << 4;
}

void SmartBenchTest::testClear()
{
	QFETCH(int, use_smart);
	switch (use_smart)
	{
	case 0:
		{
			QList<ObjetTest*> c_ptr_list;
			int nbi = iterations (QTest::currentTestFunction (), 0);
			for(int i = 0; i < nbi; i++)
			{
				c_ptr_list.append (new ObjetTest());
			}

			QBENCHMARK
			{
				if (c_ptr_list.isEmpty ()) QSKIP("Not enough data", SkipAll);
				delete c_ptr_list.at(0);
				c_ptr_list.removeFirst ();
			}
		}
		break;
	case 1:
		{
			QList< QSharedPointer<ObjetTest> > smart_ptr_list;
			int nbi = iterations (QTest::currentTestFunction (), 1);
			for(int i = 0; i < nbi; i++)
			{
				smart_ptr_list.append ( QSharedPointer<ObjetTest>(new ObjetTest()) );
			}

			QBENCHMARK
			{
				if (smart_ptr_list.isEmpty ()) QSKIP("Not enough data", SkipAll);
				smart_ptr_list.removeFirst ();
			}
		}
		break;
	case 2:
		{
			QList< QSharedPointer<ObjetTest> > pool_ptr_list;
			int nbi = iterations (QTest::currentTestFunction (), 2);
			for(int i = 0; i < nbi; i++)
			{
				pool_ptr_list.append ( createFromPool ());
			}

			QBENCHMARK
			{
				if (pool_ptr_list.isEmpty ()) QSKIP("Not enough data", SkipAll);
				pool_ptr_list.removeFirst ();
			}
		}

		qDeleteAll(_queue);
		_queue.clear ();
		break;
	case 3:
		{
			QList< std::tr1::shared_ptr<ObjetTest> > boost_ptr_list;
			int nbi = iterations (QTest::currentTestFunction (), 1);
			for(int i = 0; i < nbi; i++)
			{
				boost_ptr_list.append ( std::tr1::shared_ptr<ObjetTest>(new ObjetTest()) );
			}

			QBENCHMARK
			{
				if (boost_ptr_list.isEmpty ()) QSKIP("Not enough data", SkipAll);
				boost_ptr_list.removeFirst ();
			}
		}

		break;
	case 4:
		{
			QList< std::tr1::shared_ptr<ObjetTest> > pool_ptr_list;
			int nbi = iterations (QTest::currentTestFunction (), 2);
			for(int i = 0; i < nbi; i++)
			{
				pool_ptr_list.append ( createFromBoostPool ());
			}

			QBENCHMARK
			{
				if (pool_ptr_list.isEmpty ()) QSKIP("Not enough data", SkipAll);
				pool_ptr_list.removeFirst ();
			}
		}

		qDeleteAll(_queue);
		_queue.clear ();
		break;
	};
}

QTEST_MAIN(SmartBenchTest);

#include "tst_smartbenchtest.moc"
